use std::collections::HashMap;

use core::fmt::Debug as DebugTrait;

use crate::accessmasks::FieldAccessFlag;
use crate::classfile::{ JavaClassFile, AbstractTypeDescription, MethodInfo, AbstractTypeKind };
use crate::classstore::ClassStore;
use crate::jvm::Error;

#[derive(Debug)]
pub struct HeapArea {
    pub memory_capacity: usize,
    pub memory_used: usize,
    pub object_area: ObjectArea,
    pub static_area: StaticArea,
}

impl HeapArea {
    pub fn new(memory_capacity: usize) -> Self {
        HeapArea {
            memory_capacity,
            memory_used: 0,
            object_area: ObjectArea::default(),
            static_area: StaticArea::default(),
        }
    }

    pub fn make_object(&mut self, class_store: &ClassStore, class_index: usize) -> ObjectReference {
        let (object_ref, object_size) = self.object_area.make(class_store, class_index);
        self.memory_used += object_size;

        return object_ref;
    }

    pub fn make_static(&mut self, class: &JavaClassFile, class_index: usize) {
        self.memory_used += self.static_area.make(class, class_index);
    }
}

pub type ObjectReference=u32;

const DEFAULT_COMPARTMENT_CAPACITY: usize = u16::MAX as usize;
const INVALID_FIRST_OBJECT: usize = u16::MAX as usize;
const INVALID_NEXT_COMPARTMENT: usize = 0;

#[derive(Default, Debug)]
pub struct ObjectArea {
    compartments: Vec<ObjectCompartment>,
    first_free_compartment: usize,
}

impl ObjectArea {
    pub fn make(&mut self, class_store: &ClassStore, target_class_index: usize) -> (ObjectReference, usize) {
        let mut fields = Vec::new();
        let mut current_class_index;
        let mut next_class_index    = target_class_index;

        let mut object_size = 0;

        while next_class_index != class_store.class_count() {
            current_class_index = next_class_index;

            let class = class_store.class_file_from_idx(current_class_index).unwrap();

            for field in &class.fields {
                let new_field = ObjectField {
                    value: FieldValue::default_for(&field.descriptor)
                };
                object_size += std::mem::size_of_val(&new_field);

                fields.push(new_field);
            }

            if class.has_super_class() {
                next_class_index = class_store.class_idx_from_name(class.get_super_class_name().unwrap()).unwrap();
            } else {
                next_class_index = class_store.class_count();
            }
        }

        let new_object = HeapObject {
            class_index: target_class_index,
            fields: fields.into_boxed_slice(),
        };

        object_size += std::mem::size_of_val(&new_object);
        object_size += std::mem::size_of_val(&new_object.fields);

        let object_ref = self.store_object(new_object);

        return (object_ref, object_size);
    }

    fn store_object(&mut self, object: HeapObject) -> ObjectReference {
        if self.first_free_compartment == 0 {
            self.compartments.push(ObjectCompartment::new(INVALID_NEXT_COMPARTMENT));
            self.first_free_compartment = self.compartments.len();
        }

        let compartment_index = self.first_free_compartment - 1;
        let object_index = self.compartments.get_mut(compartment_index).unwrap().store(object);
        if self.compartments[compartment_index].first_free == INVALID_FIRST_OBJECT {
            self.first_free_compartment = self.compartments[compartment_index].next_free_compartment;
        }

        return object_index + (DEFAULT_COMPARTMENT_CAPACITY as u32 * (compartment_index+1) as u32);
    }
}

pub struct ObjectCompartment {
    objects: Box<[ObjectCompartmentEntry]>,
    first_free: usize,
    reserved_count: usize,
    next_free_compartment: usize,
}

impl ObjectCompartment {
    fn new(next_free: usize) -> Self {
        let mut os = Vec::with_capacity(DEFAULT_COMPARTMENT_CAPACITY);
        for i in 0..DEFAULT_COMPARTMENT_CAPACITY-1 {
            os.push(ObjectCompartmentEntry::EmptyNext(i+1));
        }
        os.push(ObjectCompartmentEntry::EmptyTail());

        ObjectCompartment {
            objects: os.into_boxed_slice(),
            first_free: 0,
            reserved_count: 0,
            next_free_compartment: next_free,
        }
    }

    fn store(&mut self, object: HeapObject) -> ObjectReference {
        let store_slot = self.objects.get(self.first_free).unwrap();
        let compartment_index = self.first_free;

        match store_slot {
            ObjectCompartmentEntry::Valid(_) => unreachable!(),
            ObjectCompartmentEntry::EmptyNext(next) => {
                self.first_free = *next;
            },
            ObjectCompartmentEntry::EmptyTail() => {
                // TODO: Maybe change to something else
                self.first_free = INVALID_FIRST_OBJECT;
            },
        }
        
        *self.objects.get_mut(compartment_index).unwrap() = ObjectCompartmentEntry::Valid(object);
        self.reserved_count += 1;

        return compartment_index as u32;
    }
}

impl DebugTrait for ObjectCompartment {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
        f.debug_struct("ObjectCompartment")
            .field("first_free", &self.first_free)
            .field("reserved_count", &self.reserved_count)
            .field("next_free_compartment", &self.next_free_compartment)
            .field("objects", &self.objects
                .iter()
                .enumerate()
                .filter(|e| match e { (_, ObjectCompartmentEntry::Valid(_)) => true, _ => false})
                .map(|e| match e { (i, ObjectCompartmentEntry::Valid(o)) => (i, o), _ => unreachable!()})
                .collect::<Vec<(usize, &HeapObject)>>()
            ).finish()
    }
}

#[derive(Debug)]
pub enum ObjectCompartmentEntry {
    Valid(HeapObject),
    EmptyNext(usize),
    EmptyTail(), // next empty value
}

#[derive(Debug)]
pub struct HeapObject {
    class_index: usize,
    fields: Box<[ObjectField]>,
}

#[derive(Default, Debug)]
pub struct StaticArea {
    static_objects: HashMap<String, StaticObject>,
}

impl StaticArea {

    pub fn set(&mut self, class_name: &String, field_name: &String, field_value: FieldValue) -> Result<(), Error> {
        let static_object = match self.static_objects.get_mut(class_name) {
            Some(o) => o,
            None => return Err(Error::RunTimeError(format!("Trying to set '{}.{}={:?}' but there is no such static object.", class_name, field_name, field_value))),
        };

        let field_index = match static_object.field_index_from_name(&field_name) {
            Some(i) => i,
            None => return Err(Error::RunTimeError(format!("Trying to set '{}.{}={:?}' but there is no such static field on this object.", class_name, field_name, field_value))),
        };

        field_value.check_type(&static_object.fields[field_index].descriptor, class_name)?;

        static_object.fields[field_index].value = field_value;

        Ok(())
    }

    fn make(&mut self, class: &JavaClassFile, class_index: usize) -> usize {
        let mut fields = Vec::new();

        let mut fields_cumulative_size = 0;
        for field in &class.fields {
            if field.access_flags & FieldAccessFlag::Static {
                let new_field = StaticField {
                    name: field.name.clone(),
                    descriptor: field.descriptor.clone(),
                    value: FieldValue::default_for(&field.descriptor),
                };
                fields_cumulative_size += std::mem::size_of_val(&new_field);
                fields.push(new_field);
            }
        }

        let new_object = StaticObject {
            class_index,
            fields: fields.into_boxed_slice(),
            methods: Box::new([]),
        };

        let object_memory_size = std::mem::size_of_val(&new_object);
        let field_array_size   = std::mem::size_of_val(&new_object.fields);
        let method_array_size  = std::mem::size_of_val(&new_object.methods);

        let _ = self.static_objects.insert(class.get_classname().unwrap().to_string(), new_object);

        return object_memory_size + field_array_size + fields_cumulative_size + method_array_size;
    }
}

#[derive(Debug)]
pub struct StaticObject {
    pub class_index: usize,
    pub fields: Box<[StaticField]>,
    pub methods: Box<[MethodInfo]>,
}

impl StaticObject {
    fn field_index_from_name(&self, name: &String) -> Option<usize> {
        for (index, field) in self.fields.iter().enumerate() {
            if field.name == *name {
                return Some(index)
            }
        }
        return None;
    }
}

#[derive(Debug)]
pub struct StaticField {
    pub name: String,
    pub descriptor: AbstractTypeDescription,
    pub value: FieldValue,
}

#[derive(Debug)]
pub struct ObjectField {
    pub value: FieldValue,
}

#[derive(Debug)]
pub enum FieldValue {
    Boolean(bool),
    Byte(u8),
    Char(u16),
    Short(i16),
    Int(i32),
    Float(f32),
    Reference(ObjectReference),
    Double(f64),
    Long(i64),
}

impl FieldValue {
    fn check_type(&self, t: &AbstractTypeDescription, class_name: &String) -> Result<(), Error> {
        let default_val = FieldValue::default_for(t);

        return if std::mem::discriminant(&default_val) != std::mem::discriminant(self) {
            Err(Error::RunTimeError(format!("Type Mismatch: trying to set {:?} on a field of type {:?} in class {}", self, default_val, class_name)))
        } else {
            Ok(())
        }
    }

    fn default_for(t: &AbstractTypeDescription) -> Self {
        if t.array_level != 0 {
            return Self::Reference(0);
        }
        match &t.kind {
            AbstractTypeKind::Void()       => unreachable!(),
            AbstractTypeKind::Byte()       => Self::Byte(0),
            AbstractTypeKind::Char()       => Self::Char(0),
            AbstractTypeKind::Double()     => Self::Double(0.0),
            AbstractTypeKind::Float()      => Self::Float(0.0),
            AbstractTypeKind::Int()        => Self::Int(0),
            AbstractTypeKind::Long()       => Self::Long(0),
            AbstractTypeKind::Classname(_) => Self::Reference(0),
            AbstractTypeKind::Short()      => Self::Short(0),
            AbstractTypeKind::Boolean()    => Self::Boolean(true),
        }
    }
}
